//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

import SafariServices
import Alamofire

import KakaoSDKCommon
import KakaoSDKAuth
import KakaoSDKUser
import KakaoSDKCert
import KakaoSDKTalk
import KakaoSDKFriend
import KakaoSDKShare
import KakaoSDKNavi
import KakaoSDKTemplate

// MARK: - tableview helper
class Row {
    var title: String
    var action: (() -> Void)?
    var apiResult: ApiResult
    var backgroundColor: UIColor?
    init(title: String, action: (() -> Void)?, result: ApiResult = .Ready, backgroundColor:UIColor? = nil) {
        self.title = title
        self.action = action
        self.apiResult = result
        self.backgroundColor = backgroundColor
    }
}

class SectionOfRow {
    typealias Item = Row
    
    var header: String
    var items: [Item]
    init(header:String, items:[Item]) {
        self.header = header
        self.items = items
    }
    
    required convenience init(original: SectionOfRow, items: [Item]) {
        self.init(header: original.header, items: original.items)
        self.items = items
    }
}

extension SDKSampleViewController {
    private func updateResultIndicator(cell:UITableViewCell?, row:Row) {
        
        switch(row.apiResult) {
        case .Ready: cell?.backgroundColor = UIColor.white
        case .Success: cell?.backgroundColor = UIColor.green
        case .Fail: cell?.backgroundColor = UIColor.red
        case .NotSuccess: cell?.backgroundColor = UIColor.red
        }
    }
    
    private func updateResultIndicator() {
        if let resultIndicator = self.resultIndicator {
            let cell = self.tableView.cellForRow(at: resultIndicator.0)
            self.updateResultIndicator(cell: cell, row: resultIndicator.1)
        }
    }
    
    private func success(_ object: Any? = nil, updateIndicator: Bool = true, withAlert: Bool = false, alertTitle:String? = nil, apiResult: ApiResult? = nil) {
        if let object = object {
            print("\(String(describing:object))")
            
            if withAlert == true {
                self.showAlert(title:alertTitle ?? "Success", message: String(describing: object))
            }
        }
        else {
            if withAlert == true {
                self.showAlert(title:alertTitle ?? "Success", message: "nil")
            }
        }
        
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = apiResult ?? .Success
                self.updateResultIndicator()
            }
        }
    }
    
    private func errorHandler(error:Error?, updateIndicator: Bool = true, withAlert: Bool = false) {
        guard error != nil else { return }
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = .Fail
                self.updateResultIndicator()
            }
        }
        
        if let sdkError = error as? SdkError {
            print("\n\n[sdk error handling start]------------------------------------------------------------------------------")
            
            switch(sdkError) {
            case .ApiFailed:
                let apiError = sdkError.getApiError()
                switch(apiError.reason) {
                //if need, use reason case
                default:
                    print("api error: \(String(describing: apiError))")
                }
            case .AuthFailed:
                let authError = sdkError.getAuthError()
                switch(authError.reason) {
                //if need, use reason case
                default:
                    print("auth error: \(String(describing: authError))")
                }
            case .ClientFailed:
                let clientError = sdkError.getClientError()
                switch(clientError.reason) {
                //if need, use reason case
                default:
                    print("client error: \(String(describing: clientError))")
                }
            case .AppsFailed:
                let appsError = sdkError.getAppsError()
                switch(appsError.reason) {
                //if need, use reason case
                default:
                    print("apps error: \(String(describing: appsError))")
                }
            @unknown default:
                break
            }
            
            if withAlert == true {
                self.showAlert(title:"SDK Error", message: String(describing: error))
            }
            print("[sdk error handling end]--------------------------------------------------------------------------------\n")
        }
        else {
            print("\n\n[not sdk error start]------------------------------------------------------------------------------")
            if let nsError = error as NSError? {
                print("normal error: \(String(describing: nsError))")
                
                if withAlert == true {
                    self.showAlert(title:"Normal(Not SDK) Error", message: String(describing: nsError))
                }
            }
            
            print("[not sdk error end]------------------------------------------------------------------------------\n")
        }
    }
}

// MARK: - vc
class SDKSampleViewController: SDKSampleBaseViewController, UITableViewDelegate, UITableViewDataSource {
    
//    lazy var dataSource = RxTableViewSectionedReloadDataSource<SectionOfRow>(
//        configureCell: { (dataSource, tableView, indexPath, item) in
//            let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
//
//            self.updateResultIndicator(cell:cell, row:item)
//
////            cell.selectionStyle = UITableViewCell.SelectionStyle.none
//            cell.textLabel?.adjustsFontSizeToFitWidth = true
//            cell.textLabel?.text = "\(item.title)"
//            cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
//            cell.textLabel?.textColor = UIColor.black
//            return cell
//        })
    
    var safariViewController : SFSafariViewController?
    let paramDialogRequestColor : UIColor = UIColor.init(red: 0, green: 1, blue: 1, alpha: 0.5)
    
    override func loadView() {
        super.loadView()
        
        setupViews()
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        self.tableView.delegate = self
        self.tableView.dataSource = self
        
        setMenu()
    }
    
    override func setupViews() {
        super.setupViews()
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "Cell")
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let item = self.sections![indexPath.section].items[indexPath.row]
        if let action = item.action {
            action()
            
            self.resultIndicator = (indexPath, item)
            
            DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
               tableView.deselectRow(at: indexPath, animated: false)
            }
        }
    }
    
    func numberOfSections(in tableView: UITableView) -> Int {
        if let sections = self.sections, sections.count != 0 {
//            print("sections count: ", sections.count)
            return sections.count
        }
        else {
            return 0
        }
    }
    
    func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        if let sections = self.sections, sections.count != 0 {
            return sections[section].header
        }
        else {
            return nil
        }
    }
    
    func tableView(_ tableView: UITableView, willDisplayHeaderView view: UIView, forSection section: Int) {
        (view as! UITableViewHeaderFooterView).contentView.backgroundColor = .lightGray
        (view as! UITableViewHeaderFooterView).textLabel?.textColor = .black
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if let sections = self.sections, sections.count != 0 {
            return sections[section].items.count
        }
        else {
            return 0
        }
    }

    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
        
        let item = self.sections![indexPath.section].items[indexPath.row]
        
        self.updateResultIndicator(cell:cell, row:item)
        cell.textLabel?.adjustsFontSizeToFitWidth = true
        cell.textLabel?.text = "\(item.title)"
        cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
        cell.textLabel?.textColor = UIColor.black
        
        if let backgroundColor = item.backgroundColor {
            cell.backgroundColor = backgroundColor
        }
        
        return cell
    }
    
    func setTitle(title:String = "SDK Sample", args: String? = nil) {
        self.navigationItem.title = (args != nil) ? "\(title)(\(args!))" : "\(title)"
    }
    
    func extractPickerElements(friends:Friends<Friend>) -> [(URL?, String?, Any)]? {
        return friends.elements?.map({ (friend) -> (URL?, String?, Any) in
            return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid)
        })
    }
    
    // MARK: - menu setting
    func setMenu() {
        self.sections = [
            // MARK: - section auth
            SectionOfRow(header: "AUTH", items: [
                Row(title: "hasToken() usage", action: {
                    if (AuthApi.hasToken()) {
                        UserApi.shared.accessTokenInfo { [weak self] (_, error) in
                            if let error = error {
                                if let sdkError = error as? SdkError, sdkError.isInvalidTokenError() == true  {
                                    print("login 필요")
                                    self?.errorHandler(error: error)
                                }
                                else {
                                    self?.errorHandler(error: error)
                                }
                            }
                            else {
                                self?.success("토큰 유효성 체크 성공(필요시 토큰갱신됨)")
                            }
                        }
                    }
                    else {
                        print("토큰없음: login 필요")
                        self.errorHandler(error: SdkError(reason: .TokenNotFound))
                    }
                }),
                Row(title: "refreshToken()", action: {
                    AuthApi.shared.refreshToken { [weak self](oauthToken, error) in
                        if let error = error  {
                            self?.errorHandler(error: error)
                        }
                        else  {
                            self?.success()
                        }
                    }
                }),
                
                Row(title: "getToken()", action: {
                    print("accessToken: \(Auth.shared.tokenManager.getToken()?.accessToken)")
                }),
            ]),
            // MARK: - section user login
            SectionOfRow(header: "USER - login (OIDC test는 + 항목사용)", items: [
                Row(title: "+loginWithKakaoTalk(...)", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        self.showParamDialog(apiKey: ApiKey.loginWithKakaoTalk) { paramItems in
                            UserApi.shared.loginWithKakaoTalk(launchMethod: paramItems.requestParam(key:"launchMethod"),
                                                              channelPublicIds: paramItems.requestParam(key:"channelPublicIds"),
                                                              serviceTerms: paramItems.requestParam(key:"serviceTerms"),
                                                              nonce: paramItems.requestParam(key:"nonce"),
                                                              completion: { [weak self](oauthToken, error) in
                                    if let error = error {
                                        self?.errorHandler(error: error)
                                    }
                                    else {
                                        self?.success()

                                        //do something
                                        _ = oauthToken
                                    }
                            })
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+loginWithKakaoAccount(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.loginWithKakaoAccount) { paramItems in
                        UserApi.shared.loginWithKakaoAccount(prompts:paramItems.requestParam(key:"prompts"),
                                                             loginHint: paramItems.requestParam(key:"loginHint"),
                                                             nonce: paramItems.requestParam(key:"nonce"),
                                                             completion:{ [weak self](oauthToken, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                _ = oauthToken
                                 
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+loginWithKakaoAccount(scopes:...)", action: {
                    self.showParamDialog(apiKey: ApiKey.loginWithKakaoAccountScopes) { paramItems in
                        UserApi.shared.loginWithKakaoAccount(scopes: paramItems.requestParam(key:"scopes") ?? [],
                                                             nonce: paramItems.requestParam(key:"nonce"),
                                                             completion:{ [weak self](oauthToken, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                _ = oauthToken
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "loginWithKakaoTalk()", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        UserApi.shared.loginWithKakaoTalk { [weak self](oauthToken, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                _ = oauthToken
                            }
                        }
                    }
                }),
                Row(title: "loginWithKakaoTalk(launchMethod:) - UniversalLink", action: {
                    UserApi.shared.loginWithKakaoTalk(launchMethod:.UniversalLink) { [weak self](oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = oauthToken
                        }
                    }
                }),
                Row(title: "loginWithKakaoAccount()", action: {
                    UserApi.shared.loginWithKakaoAccount { [weak self](oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = oauthToken
                             
                        }
                    }
                })
            ]),
            
            // MARK: - section user OIDC login
            SectionOfRow(header: "USER - OIDC login", items: [
                Row(title: "OIDC me() incremental authorization required", action: {
                    UserApi.shared.me() { [weak self] (user, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            
                            if let user = user {
                                var scopes = [String]()
                                if (user.kakaoAccount?.emailNeedsAgreement == true) { scopes.append("account_email") }
                                if (user.kakaoAccount?.birthdayNeedsAgreement == true) { scopes.append("birthday") }
                                if (user.kakaoAccount?.birthyearNeedsAgreement == true) { scopes.append("birthyear") }
                                if (user.kakaoAccount?.ciNeedsAgreement == true) { scopes.append("account_ci") }
                                if (user.kakaoAccount?.phoneNumberNeedsAgreement == true) { scopes.append("phone_number") }
                                if (user.kakaoAccount?.profileNeedsAgreement == true) { scopes.append("profile") }
                                if (user.kakaoAccount?.ageRangeNeedsAgreement == true) { scopes.append("age_range") }
                                
                                if scopes.count > 0 {
                                    //OpenID 활성화 후
                                    // - scope 파라메터에 openid 항목을 포함하여 요청할 경우 OIDC로 동작
                                    // - scope 파라메터에 openid 항목을 미포함 시 OAuth2.0 으로 동작
                                    
                                    // OIDC 요청이므로 "openid" 항목을 추가한다
                                    scopes.append("openid")
                                    
                                    //필요한 scope으로 토큰갱신을 한다.
                                    UserApi.shared.loginWithKakaoAccount(scopes: scopes, nonce:"test_nonce") { (_, error) in
                                        if let error = error {
                                            self?.errorHandler(error: error)
                                        }
                                        else {
                                            UserApi.shared.me() { (user, error) in
                                                if let error = error {
                                                    self?.errorHandler(error: error)
                                                }
                                                else {
                                                    self?.success()
                                                    
                                                    //do something.
                                                    _ = user
                                                }
                                            }
                                        }
                                    }
                                }
                                else {
                                    self?.success("--> 추가 동의가 필요하지 않다.")
                                }
                            }
                        }
                    }
                }),
            ]),
            // MARK: - section user other api
            SectionOfRow(header: "USER - other api", items: [
                Row(title: "signup()", action: {
                    UserApi.shared.signup { [weak self] (userId, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title:"logout()", action: {
                    UserApi.shared.logout { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title:"unlink()", action: {
                    let alert = UIAlertController(title: "unlink()", message: "are you sure?", preferredStyle: UIAlertController.Style.alert)
                    let ok = UIAlertAction(title: "OK", style: .default, handler: { action in
                        UserApi.shared.unlink { [weak self] (error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                            }
                        }
                    })
                    alert.addAction(ok)
                    let cancel = UIAlertAction(title: "Cancel", style: .default, handler: nil)
                    alert.addAction(cancel)
                    DispatchQueue.main.async(execute: {
                        self.present(alert, animated: true)
                    })
                }),
                Row(title: "+me(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.me) { paramItems in
                        UserApi.shared.me(propertyKeys: paramItems.requestParam(key:"propertyKeys"),
                                          completion: { [weak self] (user, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(user)
                                
                                //do something
                                _ = user
                                
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "me() incremental authorization required", action: {
                    UserApi.shared.me() { [weak self] (user, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            
                            if let user = user {
                                var scopes = [String]()
                                if (user.kakaoAccount?.emailNeedsAgreement == true) { scopes.append("account_email") }
                                if (user.kakaoAccount?.birthdayNeedsAgreement == true) { scopes.append("birthday") }
                                if (user.kakaoAccount?.birthyearNeedsAgreement == true) { scopes.append("birthyear") }
                                if (user.kakaoAccount?.ciNeedsAgreement == true) { scopes.append("account_ci") }
                                if (user.kakaoAccount?.phoneNumberNeedsAgreement == true) { scopes.append("phone_number") }
                                if (user.kakaoAccount?.profileNeedsAgreement == true) { scopes.append("profile") }
                                if (user.kakaoAccount?.ageRangeNeedsAgreement == true) { scopes.append("age_range") }
                                
                                if scopes.count > 0 {
                                    //필요한 scope으로 토큰갱신을 한다.
                                    UserApi.shared.loginWithKakaoAccount(scopes: scopes) { (_, error) in
                                        if let error = error {
                                            self?.errorHandler(error: error)
                                        }
                                        else {
                                            UserApi.shared.me() { (user, error) in
                                                if let error = error {
                                                    self?.errorHandler(error: error)
                                                }
                                                else {
                                                    self?.success()
                                                    
                                                    //do something.
                                                    _ = user
                                                }
                                            }
                                        }
                                    }
                                }
                                else {
                                    self?.success("--> 추가 동의가 필요하지 않다.")
                                }
                            }
                        }
                    }
                }),
                Row(title:"updateProfile()", action: {
                    UserApi.shared.updateProfile(properties: ["custom_key":"update_profile_test"]) { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                        }
                    }
                }),
                Row(title:"accessTokenInfo()", action: {
                    UserApi.shared.accessTokenInfo { [weak self] (accessTokenInfo, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = accessTokenInfo
                            
                        }
                    }
                }),
                Row(title:"shippingAddresses()", action: {
                    UserApi.shared.shippingAddresses(completion: { [weak self] (shippingAddress, error) in

                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = shippingAddress
                        }
                    })
                }),
                Row(title:"shippingAddresses(fromUpdatedAt:pageSize:)", action: {
                    UserApi.shared.shippingAddresses(fromUpdatedAt: Date(), completion: { [weak self] (shippingAddress, error) in

                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = shippingAddress
                        }
                    })
                }),
                Row(title:"+serviceTerms(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.serviceTerms) { paramItems in
                        UserApi.shared.serviceTerms(result:paramItems.requestParam(key:"result"), tags: paramItems.requestParam(key: "tags"),
                                                    completion:{ [weak self] (serviceTerms, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(serviceTerms)
                                
                                //do something
                                _ = serviceTerms
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                
                Row(title: "+revokeServiceTerms(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.revokeServiceTerms) { paramItems in
                        UserApi.shared.revokeServiceTerms(tags: paramItems.requestParam(key: "serviceTerms") ?? []) {[weak self] (revokes, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            } else {
                                self?.success(revokes)                                
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                
                Row(title: "+scopes(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.scopes) { paramItems in
                        UserApi.shared.scopes(scopes: paramItems.requestParam(key:"scopes") ?? [],
                                              completion: { [weak self] (scopeInfo, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(scopeInfo)
                                
                                //do something
                                _ = scopeInfo
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+revokeScopes(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.revokeScopes) { paramItems in
                        UserApi.shared.revokeScopes(scopes: paramItems.requestParam(key:"scopes") ?? [],
                                                    completion: { [weak self] (scopeInfo, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(scopeInfo)
                                
                                //do something
                                _ = scopeInfo
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                
                Row(title:"selectShippingAddress()", action: {
                    UserApi.shared.selectShippingAddress { [weak self] (addressId, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        } else {
                            self?.success(addressId)
                        }
                    }
                })
            ]),
            // MARK: - section cert 인증 login
            SectionOfRow(header: "CERT - (k2100) 인증 login (OIDC test는 + 항목사용)", items: [
                Row(title: "+certLoginWithKakaoTalk(...)", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        guard let apiParams = self.apiParams(ApiKey.certLoginWithKakaoTalk) else { return }
                        let vc = SDKSampleParamViewController(apiParams: apiParams,
                                                              paramViewCompletionHandler: { [weak self] (resultApiParams) in

                            self?.apiParamsInfo.info[ApiKey.certLoginWithKakaoTalk] = resultApiParams
                            self?.saveApiParamsInfo()

                            let paramItems = resultApiParams.paramItems
                            CertApi.shared.certLoginWithKakaoTalk(certType: paramItems.requestParam(key: "certType") ?? .K2100,
                                                                  launchMethod: paramItems.requestParam(key:"launchMethod"),
                                                                  prompts: paramItems.requestParam(key:"prompts"),
                                                                  channelPublicIds: paramItems.requestParam(key:"channelPublicIds"),
                                                                  serviceTerms: paramItems.requestParam(key:"serviceTerms"),
                                                                  signData: paramItems.requestParam(key:"signData"),
                                                                  nonce: paramItems.requestParam(key:"nonce"),
                                                                  settleId: paramItems.requestParam(key:"settleId"),
                                                                  identifyItems: paramItems.requestParam(key: "identifyItems"),
                                                                  completion: { [weak self](certTokenInfo, error) in
                                if let error = error {
                                    self?.errorHandler(error: error)
                                }
                                else {
                                    self?.success(certTokenInfo)

                                    //do something
                                    _ = certTokenInfo?.token
                                    _ = certTokenInfo?.txId

                                }
                            })
                        })
                        self.presentPopup(vc: vc)
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+certLoginWithKakaoAccount(...)", action: {
                    guard let apiParams = self.apiParams(ApiKey.certLoginWithKakaoAccount) else { return }
                    let vc = SDKSampleParamViewController(apiParams: apiParams,
                                                          paramViewCompletionHandler: { [weak self] (resultApiParams) in

                        self?.apiParamsInfo.info[ApiKey.certLoginWithKakaoAccount] = resultApiParams
                        self?.saveApiParamsInfo()

                        let paramItems = resultApiParams.paramItems
                        CertApi.shared.certLoginWithKakaoAccount(certType: paramItems.requestParam(key: "certType") ?? .K2100,
                                                                 prompts:paramItems.requestParam(key:"prompts"),
                                                                 loginHint: paramItems.requestParam(key:"loginHint"),
                                                                 signData: paramItems.requestParam(key:"signData"),
                                                                 nonce: paramItems.requestParam(key:"nonce"),
                                                                 settleId: paramItems.requestParam(key:"settleId"),
                                                                 identifyItems: paramItems.requestParam(key: "identifyItems"),
                                                                 completion:{ certTokenInfo, error in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(certTokenInfo)

                                //do something
                                _ = certTokenInfo?.token
                                _ = certTokenInfo?.txId

                            }
                        })
                    })
                    self.presentPopup(vc: vc)
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "certLoginWithKakaoTalk(settleId:)", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        CertApi.shared.certLoginWithKakaoTalk(certType:.K2100, settleId:AppDelegate.settleId()) {[weak self](certTokenInfo, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(certTokenInfo)
                                
                                //do something
                                _ = certTokenInfo?.token
                                _ = certTokenInfo?.txId
                            }
                        }
                    }
                }),
                Row(title: "certLoginWithKakaoAccount(settleId:)", action: {
                    CertApi.shared.certLoginWithKakaoAccount(certType:.K2100, settleId:AppDelegate.settleId()) { [weak self](certTokenInfo, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(certTokenInfo)
                            
                            //do something
                            _ = certTokenInfo?.token
                            _ = certTokenInfo?.txId
                             
                        }
                    }
                }),
            ]),
            // MARK: - K2220
            SectionOfRow(header: "CERT - (k2220)", items: [
                Row(title: "TEST", action: {                    
                }),
                Row(title: "[SDK]publicKey(certType:)", action: {
                    if let publicKey = CertApi.shared.publicKey(certType:.K2220) {
                        self.success(publicKey, withAlert: true, alertTitle: "PublicKey[k2220]")
                    }
                    else {
                        self.errorHandler(error: CertDemoError(msg: "public key[k2220] error"), withAlert: true)
                    }
                }),
                Row(title: "[SDK]sessionInfo(certType:) request", action: {
                    guard let txId = CertDemoApi.shared.txId(certType:.K2220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k2220] is nil. login first."), withAlert: true)
                        return
                    }
                    
                    CertApi.shared.sessionInfo(certType:.K2220, txId: txId) { [weak self] sessionInfo, error in
                        if let error = error  {
                            self?.errorHandler(error: error, withAlert: true)
                        }
                        else {
                            self?.success(sessionInfo, withAlert: true)
                        }
                    }
                }),
                Row(title: "[SDK]sessionInfo(certType:)", action: {
                    if let sessionInfo = CertApi.shared.sessionInfo(certType:.K2220) {
                        self.success(sessionInfo, withAlert: true, alertTitle: "SessionInfo[k2220]", apiResult:.Success)
                    }
                    else {
                        self.success(nil, withAlert: true, alertTitle: "SessionInfo[k2220]", apiResult:.NotSuccess)
                    }
                }),
                Row(title: "[SDK]isValidSession(certType:)", action: {
                    let isValidSession = CertApi.shared.isValidSession(certType:.K2220)
                    if isValidSession == true {
                        self.success(isValidSession, withAlert: true, alertTitle: "IsValidSession[k2220]", apiResult:.Success)
                    }
                    else {
                        self.success(isValidSession, withAlert: true, alertTitle: "IsValidSession[k2220]", apiResult:.NotSuccess)
                    }
                }),
                Row(title: "[SDK]deleteKeyPair(certType:)", action: {
                    let alert = UIAlertController(title: "deleteKeyPair()[k2220]", message: "이전에 생성된 키쌍과 기록되어있던 세션정보는 삭제됩니다. are you sure?", preferredStyle: UIAlertController.Style.alert)
                    let ok = UIAlertAction(title: "OK", style: .default, handler: { action in
                        CertApi.shared.deleteKeyPair(certType:.K2220)
                        
                        //[이용기관]도 txId를 삭제해야한다.
                        CertDemoApi.shared.deleteTxId(certType:.K2220)
                        
//                        self.success("이전에 생성된 키쌍과 기록되어있던 세션정보는 삭제되었으며\n이후 publicKey()를 호출하면 키쌍은 재생성 됩니다.", withAlert: true, alertTitle: "DeleteKeyPair")
                        self.success()
                    })
                    alert.addAction(ok)
                    let cancel = UIAlertAction(title: "Cancel", style: .default, handler: nil)
                    alert.addAction(cancel)
                    DispatchQueue.main.async(execute: {
                        self.present(alert, animated: true)
                    })
                }),
                Row(title: "[이용기관]demoSign(...) -> [SDK]certLoginWithKakaoTalk(..)", action: {
                    if (UserApi.isKakaoTalkLoginAvailable()) {
                        guard let publicKey = CertApi.shared.publicKey(certType:.K2220) else {
                            self.errorHandler(error: CertDemoError(msg: "public key[k2220] is nil"), withAlert: true)
                            return
                        }
                        
                        CertDemoApi.shared.demoSign(certType:.K2220,
                                                    base64encodedDerPublicKey:publicKey,
                                                    completion: { [weak self] (txId, error) in
                            if let error = error  {
                                self?.errorHandler(error: error, withAlert: true)
                            }
                            else  {
                                if let txId = txId {
                                    CertApi.shared.certLoginWithKakaoTalk(certType: .K2220, txId: txId, settleId: AppDelegate.settleId()) { certTokenInfo, error in
                                        if let error = error {
                                            if let sdkError = error as? SdkError {
                                                //실패시 키생성->txId발급->로그인을 다시 해야하므로 deleteKeyPair(), deleteTxId()를 호출한다.
                                                CertApi.shared.deleteKeyPair(certType:.K2220)
                                                CertDemoApi.shared.deleteTxId(certType:.K2220)
                                                self?.errorHandler(error: sdkError, withAlert: true)
                                            }
                                            else {
                                                self?.errorHandler(error: error, withAlert: true)
                                            }
                                        }
                                        else  {
                                            CertDemoApi.shared.saveTxId(txId, certType:.K2220)
                                            self?.success(certTokenInfo, withAlert: true)
                                        }
                                    }
                                }
                                else {
                                    //발생하면 안되는 케이스..
                                    //만약 발생시 키생성->txId발급->로그인을 다시 해야하므로 deleteKeyPair(), deleteTxId()를 호출한다
                                    CertApi.shared.deleteKeyPair(certType:.K2220)
                                    CertDemoApi.shared.deleteTxId(certType:.K2220)
                                    self?.errorHandler(error: CertDemoError(msg: "txId[k2220] is nil"), withAlert: true)
                                }
                            }
                        })
                    }
                    else {
                        self.errorHandler(error: CertDemoError(msg: "kakaotalk is not available"), withAlert: true)
                        
                        //예) 톡설치용 유니버셜 링크 호출 예
                        //UserApi.shared.loginWithKakaoTalk(launchMethod: .UniversalLink) { }
                        return
                    }
                }),
                Row(title: "[이용기관]demoVerify(txId:appUserId:certType:)", action: {
                    guard let txId = CertDemoApi.shared.txId(certType:.K2220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k2220] is nil. login first."), withAlert: true)
                        return
                    }
                    
                    UserApi.shared.me { user, error in
                        if let error = error {
                            self.errorHandler(error:error, withAlert: true)
                        }
                        else {
                            guard let userId = user?.id else {
                                self.errorHandler(error: CertDemoError(msg: "[k2220]user.id is nil"), withAlert: true)
                                return
                            }
                            
                            CertDemoApi.shared.demoVerify(certType:.K2220, txId: txId, appUserId: userId) { [weak self] response, error in
                                if let error = error {
                                    self?.errorHandler(error:error, withAlert: true)
                                }
                                else {
                                    self?.success(response, withAlert: true)
                                }
                            }
                        }
                    }
                }),
                Row(title: "[이용기관]demoReducedSign(txId:data:signature:certType:)", action: {
                    guard let txId = CertDemoApi.shared.txId(certType:.K2220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k2220] is nil. login first."), withAlert: true)
                        return
                    }
                    
                    let signData = "hello"
                    
                    CertApi.shared.reducedSign(certType:.K2220, data: signData, completion: { [weak self]signature, error in
                        if let error = error {
                            self?.errorHandler(error: error, withAlert: true)
                        }
                        else  {
                            guard let signature = signature else {
                                self?.errorHandler(error: CertDemoError(msg:"[k2220]signature is nil"), withAlert: true)
                                return
                            }
                            
                            CertDemoApi.shared.demoReducedSign(certType:.K2220, txId: txId, data: signData, signature: signature, completion:{ response, error in
                                if let error = error  {
                                    self?.errorHandler(error: error, withAlert: true)
                                }
                                else  {
                                    self?.success(response, withAlert: true)
                                }
                            })
                        }
                    })
                }),
//                Row(title: "[이용기관]demoSignTest", action: {
//
//                    CertDemoApi.shared.demoSignTest(digest: "test", completion:{ [weak self] _, error in
//                        if let error = error  {
//                            self?.errorHandler(error: error, withAlert: true)
//                        }
//                        else  {
//                            self?.success(withAlert: true)
//                        }
//                    })
//                }),
//                Row(title: "delete_test_keychain", action: {
//                    KeyChainAccessGroupHelper.deleteAccessGroupInfo()
//                })
            ]),
            // MARK: - k3220
            SectionOfRow(header: "CERT - (k3220)", items: [
                Row(title: "[SDK]publicKey(certType:)", action: {
                    if let publicKey = CertApi.shared.publicKey(certType:.K3220) {
                        self.success(publicKey, withAlert: true, alertTitle: "PublicKey[k3220]")
                    }
                    else {
                        self.errorHandler(error: CertDemoError(msg: "public key[k3220] error"), withAlert: true)
                    }
                }),
                Row(title: "[SDK]sessionInfoByAppKey(certType:txId:targetAppKey:) request", action: {
                    let targetAppKey = CertDemoApi.shared.demoAppKey.appKeys()[0]
                    
                    guard let txId = CertDemoApi.shared.txId(certType:.K3220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k3220] is nil. sign first."), withAlert: true)
                        return
                    }
                    
                    CertApi.shared.sessionInfoByAppKey(certType:.K3220, txId: txId, targetAppKey:targetAppKey) { [weak self] sessionInfo, error in
                        if let error = error  {
                            self?.errorHandler(error: error, withAlert: true)
                        }
                        else {
                            self?.success(sessionInfo, withAlert: true, alertTitle: "sessionInfo[k3220]")
                        }
                    }
                }),
                Row(title: "[SDK]sessionInfo(certType:)", action: {
                    if let sessionInfo = CertApi.shared.sessionInfo(certType:.K3220) {
                        self.success(sessionInfo, withAlert: true, alertTitle: "SessionInfo[k3220]", apiResult:.Success)
                    }
                    else {
                        self.success(nil, withAlert: true, alertTitle: "SessionInfo[k3220]", apiResult:.NotSuccess)
                    }
                }),
                Row(title: "[SDK]isValidSession(certType:)", action: {
                    let isValidSession = CertApi.shared.isValidSession(certType:.K3220)
                    if isValidSession == true {
                        self.success(isValidSession, withAlert: true, alertTitle: "IsValidSession[k3220]", apiResult:.Success)
                    }
                    else {
                        self.success(isValidSession, withAlert: true, alertTitle: "IsValidSession[k3220]", apiResult:.NotSuccess)
                    }
                }),
                Row(title: "[SDK]deleteKeyPair(certType:)", action: {
                    let alert = UIAlertController(title: "deleteKeyPair(certType:)[k3220]", message: "이전에 생성된 키쌍과 기록되어있던 세션정보는 삭제됩니다. are you sure?", preferredStyle: UIAlertController.Style.alert)
                    let ok = UIAlertAction(title: "OK", style: .default, handler: { action in
                        CertApi.shared.deleteKeyPair(certType:.K3220)
                        
                        //[이용기관]도 txId를 삭제해야한다.
                        CertDemoApi.shared.deleteTxId(certType:.K3220)
                        
//                        self.success("이전에 생성된 키쌍과 기록되어있던 세션정보는 삭제되었으며\n이후 publicKey()를 호출하면 키쌍은 재생성 됩니다.", withAlert: true, alertTitle: "DeleteKeyPair")
                        self.success()
                    })
                    alert.addAction(ok)
                    let cancel = UIAlertAction(title: "Cancel", style: .default, handler: nil)
                    alert.addAction(cancel)
                    DispatchQueue.main.async(execute: {
                        self.present(alert, animated: true)
                    })
                }),
                Row(title: "[이용기관]demoSign(...) -> [SDK]signWithKakaoTalk(..)", action: {
                    let returnUrl = CertApi.certReturnUrl()
                    let targetAppKey = CertDemoApi.shared.demoAppKey.appKeys()[0]
                    
                    if (CertApi.isKakaoTalkSignAvailable()) {
                        guard let publicKey = CertApi.shared.publicKey(certType: .K3220) else {
                            self.errorHandler(error: CertDemoError(msg: "public key[k3220] is nil"), withAlert: true)
                            return
                        }
                        
                        CertDemoApi.shared.demoSign(certType: .K3220,
                                                    base64encodedDerPublicKey:publicKey,
                                                    returnUrl: returnUrl,
                                                    validateAppKey: try! KakaoSDK.shared.appKey(),
                                                    targetAppKey: targetAppKey,
                                                    completion: { [weak self] (txId, error) in
                            if let error = error  {
                                self?.errorHandler(error: error, withAlert: true)
                            }
                            else  {
                                if let txId = txId {
                                    CertApi.shared.signWithKakaoTalk(certType:.K3220, txId: txId, targetAppKey: targetAppKey) { signStatusInfo, error in
                                        if let error = error {
                                            if let sdkError = error as? SdkError {
                                                //실패시 서명을 다시 해야하므로 deleteKeyPair(), deleteTxId()를 호출한다.
                                                CertApi.shared.deleteKeyPair(certType: .K3220)
                                                CertDemoApi.shared.deleteTxId(certType: .K3220)
                                                self?.errorHandler(error: sdkError, withAlert: true)
                                            }
                                            else {
                                                self?.errorHandler(error: error, withAlert: true)
                                            }
                                        }
                                        else {
                                            if let txId = signStatusInfo?.txId {
                                                CertDemoApi.shared.saveTxId(txId, certType: .K3220)
                                                self?.success(txId, withAlert: true)
                                            }
                                        }
                                    }
                                }
                                else {
                                    //발생하면 안되는 케이스..
                                    //만약 발생시 서명을 다시해야하므로 deleteKeyPair(), deleteTxId()를 호출한다
                                    CertApi.shared.deleteKeyPair(certType: .K3220)
                                    CertDemoApi.shared.deleteTxId(certType: .K3220)
                                    self?.errorHandler(error: CertDemoError(msg: "txId[k3220] is nil"), withAlert: true)
                                }
                            }
                        })
                    }
                    else {
                        self.errorHandler(error: CertDemoError(msg: "kakaotalk is not available"), withAlert: true)
                        
                        //예) 톡설치용 유니버셜 링크 호출 예
                        //UserApi.shared.loginWithKakaoTalk(launchMethod: .UniversalLink) { }
                        return
                    }
                }),
                Row(title: "[이용기관]demoVerify(txId:targetAppKey:certType:)", action: {
                    let targetAppKey = CertDemoApi.shared.demoAppKey.appKeys()[0]
                    
                    guard let txId = CertDemoApi.shared.txId(certType:.K3220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k3220] is nil. sign first."), withAlert: true)
                        return
                    }
                    
                    CertDemoApi.shared.demoVerify(certType: .K3220, txId: txId, targetAppKey: targetAppKey) { [weak self] response, error in
                        if let error = error {
                            self?.errorHandler(error:error, withAlert: true)
                        }
                        else {
                            self?.success(response, withAlert: true)
                        }
                    }
                }),
                Row(title: "[이용기관]demoReducedSign(txId:data:signature:certType:)", action: {
                    guard let txId = CertDemoApi.shared.txId(certType:.K3220) else {
                        self.errorHandler(error: CertDemoError(msg: "txId[k3220] is nil. sign first."), withAlert: true)
                        return
                    }
                    
                    let signData = "hello"
                    
                    CertApi.shared.reducedSign(certType:.K3220, data: signData, completion: { [weak self]signature, error in
                        if let error = error {
                            self?.errorHandler(error: error, withAlert: true)
                        }
                        else  {
                            guard let signature = signature else {
                                self?.errorHandler(error: CertDemoError(msg:"[k3220]signature is nil"), withAlert: true)
                                return
                            }
                            
                            CertDemoApi.shared.demoReducedSign(certType:.K3220, txId: txId, data: signData, signature: signature, completion:{ response, error in
                                if let error = error  {
                                    self?.errorHandler(error: error, withAlert: true)
                                }
                                else  {
                                    self?.success(response, withAlert: true)
                                }
                            })
                        }
                    })
                }),
            ]),
            // MARK: - section talk
            SectionOfRow(header: "TALK", items: [
                Row(title: "profile()", action: {
                    TalkApi.shared.profile { [weak self] (profile, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(profile)
                            
                            //do something
                            _ = profile
                        }
                    }
                }),
                Row(title: "friends()", action: {
                    TalkApi.shared.friends { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends)
                            
                            //do something
                            _ = friends
                        }
                    }
                }),
                Row(title: "+friends(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.friends) { paramItems in
                        TalkApi.shared.friends(offset: paramItems.requestParam(key:"offset"),
                                               limit: paramItems.requestParam(key:"limit"),
                                               order: paramItems.requestParam(key:"order"),
                                               friendOrder: paramItems.requestParam(key:"friendOrder"),
                                               completion: { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends)
                                
                                //do something
                                _ = friends
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "friends(context:) - recursive", action: {
                    var nextFriendContext : FriendsContext? = FriendsContext(offset: 0, limit: 3, order: .Desc)                    
                    self.recursiveAppFriendsCompletion = { [weak self] (friends, error) -> Void in
                        guard let strongSelf = self else { return }

                        if let friends = friends  {
                            guard let next = FriendsContext(friends.afterUrl) else { return }
//                            guard let prev = FriendsContext(friends.beforeUrl) else { return }
                            nextFriendContext = next
                        }

                        TalkApi.shared.friends(context: nextFriendContext, completion: { (friends, error) in
                            if let error = error {
                                strongSelf.errorHandler(error: error)
                            }
                            else {

                                strongSelf.success()
                                strongSelf.recursiveAppFriendsCompletion?(friends, nil)
                            }
                        })
                    }

                    self.recursiveAppFriendsCompletion?(nil, nil)
                }),
                Row(title: "friends(context:) - FriendsContext", action: {
                    let nextFriendContext : FriendsContext? = FriendsContext(offset:3, limit:3, order:.Desc)
                    
                    //first
                    TalkApi.shared.friends(context:nextFriendContext) { (friends, error) in
                        if let error = error {
                            self.errorHandler(error: error)
                        }
                        else {
                            self.success()
                            _ = friends
                        }
                    }
                }),
                Row(title: "friends(offset:limit:order:)", action: {
                    TalkApi.shared.friends(offset:0, limit:1, order: .Desc) { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends)
                            
                            //do something
                            _ = friends
                        }
                    }
                }),
                Row(title: "+channels(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.channels) { paramItems in
                        TalkApi.shared.channels(publicIds:paramItems.requestParam(key:"publicIds"),
                                                completion: { [weak self] (channels, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(channels)
                                
                                //do something
                                _ = channels
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+addChannel(...) - web", action: {
                    self.showParamDialog(apiKey: ApiKey.addChannel, completionDelay: 0.4) { paramItems in
                        let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                        if let channelPublicId = channelPublicId {
                            self.safariViewController
                                = SFSafariViewController(url: TalkApi.shared.makeUrlForAddChannel(channelPublicId:channelPublicId)!)
                            
                            guard (self.safariViewController != nil) else { return }
    
                            self.safariViewController?.modalTransitionStyle = .coverVertical
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("add plusfriend present success")
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+chatChannel(...) - web", action: {
                    self.showParamDialog(apiKey: ApiKey.chatChannel, completionDelay: 0.4) { paramItems in
                        let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                        if let channelPublicId = channelPublicId {
                            self.safariViewController
                                = SFSafariViewController(url: TalkApi.shared.makeUrlForChatChannel(channelPublicId:channelPublicId)!)
                            self.safariViewController?.modalTransitionStyle = .crossDissolve
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("chat present success")
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+addChannel(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.addChannel, completionDelay: 0.4) { paramItems in
                        let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                        if let channelPublicId = channelPublicId {
                            TalkApi.shared.addChannel(channelPublicId: channelPublicId) { [weak self] error in
                                if let error = error {
                                    self?.errorHandler(error: error)
                                } else {
                                    self?.success()
                                }
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+chatChannel(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.chatChannel, completionDelay: 0.4) { paramItems in
                        let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                        if let channelPublicId = channelPublicId {
                            TalkApi.shared.chatChannel(channelPublicId: channelPublicId) { [weak self] error in
                                if let error = error {
                                    self?.errorHandler(error: error)
                                } else {
                                    self?.success()
                                }
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "+followChannel(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.followChannel, completionDelay: 0.4) { paramItems in
                        let channelPublicId:String? = paramItems.requestParam(key:"channelPublicId")
                        if let channelPublicId = channelPublicId {
                            TalkApi.shared.followChannel(channelPublicId: channelPublicId) { [weak self] (result, error) in
                                if let error = error {
                                    self?.errorHandler(error: error)
                                } else {
                                    self?.success(result)
                                }
                            }
                        }
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "sendDefaultMemo(templatable:)", action: {
                    //MARK: 자기자신에게 send memo
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                      "item_content" : {
                         "profile_text" :"Kakao",
                         "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                         "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                         "title_image_text" :"Cheese cake",
                         "title_image_category" : "cake",
                         "items" : [
                            {
                                "item" :"cake1",
                                "item_op" : "1000원",
                            },
                            {
                                "item" :"cake2",
                                "item_op" : "2000원",
                            },
                            {
                                "item" :"cake3",
                                "item_op" : "3000원",
                            },
                            {
                                "item" :"cake4",
                                "item_op" : "4000원",
                            },
                            {
                                "item" :"cake5",
                                "item_op" : "5000원",
                            }
                         ],
                         "sum" :"total",
                         "sum_op" : "15000원"
                       },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    //TODO: 디코드 메소드 래핑
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        print("Template URL decoding test: \(templatable.content.imageUrl?.absoluteString)")
                        TalkApi.shared.sendDefaultMemo(templatable: templatable) { [weak self] (error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                //do something
                            }
                        }
                    }
                }),
                Row(title: "sendCustomMemo(templateId:)", action: {
                    TalkApi.shared.sendCustomMemo(templateId: AppDelegate.templateId().message) { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                        }
                    }
                }),
                Row(title: "sendCustomMemo(templateId:templateArgs:)", action: {
                    TalkApi.shared.sendCustomMemo(templateId: AppDelegate.templateId().message, templateArgs: ["msg":"메세지 파라미터 테스트입니다."]) { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                        }
                    }
                }),
                Row(title: "sendScrapMemo(requestUrl:)", action: {
                    TalkApi.shared.sendScrapMemo(requestUrl: "https://developers.kakao.com") { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                        }
                    }
                }),

                //MARK: 톡 친구에게 send message - 디폴트메세지
                Row(title: "sendDefaultMessage(templatable:receiverUuids:) - feed", action: {

                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - list", action: {

                    let listTemplateJsonStringData =
                      """
                      {
                          "object_type": "list",
                          "header_title": "WEEKLY MAGAZINE",
                          "header_link": {
                              "mobile_web_url": "https://developers.kakao.com",
                              "web_url": "https://developers.kakao.com"
                          },
                          "contents": [
                              {
                                  "title": "취미의 특징, 탁구",
                                  "description": "스포츠",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/bDPMIb/btqgeoTRQvd/49BuF1gNo6UXkdbKecx600/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "크림으로 이해하는 커피이야기",
                                  "description": "음식",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/QPeNt/btqgeSfSsCR/0QJIRuWTtkg4cYc57n8H80/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "감성이 가득한 분위기",
                                  "description": "사진",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/c7MBX4/btqgeRgWhBy/ZMLnndJFAqyUAnqu4sQHS0/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ],
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {

                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - location", action: {

                    let locationTemplateJsonStringData =
                      """
                      {
                          "object_type": "location",
                          "address_title": "카카오 판교오피스 카페톡",
                          "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 8층",
                          "content": {
                              "description": "이번 주는 체리블라썸라떼 1+1",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/bSbH9w/btqgegaEDfW/vD9KKV0hEintg6bZT4v4WK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.com"
                              },
                              "title": "신메뉴 출시❤️ 체리블라썸라떼"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          }
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }

                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - commerce", action: {

                    let commerceTemplateJsonStringData =
                      """
                      {
                          "object_type": "commerce",
                          "commerce": {
                              "product_name": "Ivory long dress",
                              "regular_price": 208800,
                              "discount_price": 146160,
                              "discount_rate": 30,
                              "currency_unit": "₩",
                              "currency_unit_position": 1
                          },
                          "content": {
                              "title": "Ivory long dress (4 Color)",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/RY8ZN/btqgOGzITp3/uCM1x2xu7GNfr7NS9QvEs0/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              },
                          },
                          "buttons": [
                              {
                                  "title": "구매하기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "공유하기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - text", action: {
                    let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                    let textTemplateJsonStringData =
                    """
                    {
                        "object_type": "text",
                        "text": "\(text)",
                        "link": {
                            "web_url": "https://developers.kakao.com",
                            "mobile_web_url": "https://developers.kakao.com"
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - calendar", action: {

                    let calendarTemplateJsonStringData =
                      """
                      {
                          "object_type": "calendar",
                          "id_type": "\(CalendarTemplate.IdType.Calendar.rawValue)",
                          "id": "\(AppDelegate.calendarId().calendar)",
                          "content": {
                              "link": {
                                  "web_url": "https://devtalk.kakao.com",
                                  "mobile_web_url": "https://developers.kakao.com"
                              },
                              "title": "2022 카타르 월드컵",
                              "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                          },
                          "buttons": [
                              {
                                  "title": "월드컵 일정 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - calendar(event)", action: {

                    let calendarTemplateJsonStringData =
                      """
                      {
                          "object_type": "calendar",
                          "id_type": "\(CalendarTemplate.IdType.Event.rawValue)",
                          "id": "\(AppDelegate.calendarId().event)",
                          "content": {
                              "link": {
                                  "web_url": "https://devtalk.kakao.com",
                                  "mobile_web_url": "https://developers.kakao.com"
                              },
                              "title": "1월 신작 평론 모임",
                              "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                          },
                          
                          "buttons": [
                              {
                                  "title": "모임 주제 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        TalkApi.shared.friends { [weak self] (friends, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(friends, updateIndicator: false)
                                guard let friends = friends else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                
                                                                                TalkApi.shared.sendDefaultMessage(templatable: templatable, receiverUuids: friendUuids) { (messageSendResult, error) in
                                                                                    if let error = error {
                                                                                        self?.errorHandler(error: error)
                                                                                    }
                                                                                    else {
                                                                                        self?.success()
                                                                                        
                                                                                        //do something
                                                                                        _ = messageSendResult
                                                                                    }
                                                                                }

                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        }
                    }
                }),
                //MARK: 톡 친구에게 send message - 커스텀메세지
                Row(title: "sendCustomMessage(templateId:templateArgs:receiverUuids:)", action: {
                    TalkApi.shared.friends { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            guard let friends = friends else {
                               self?.errorHandler(error: error)
                               return
                            }
                           
                            if let elements = self?.extractPickerElements(friends: friends) {
                                let vc = SDKSamplePickerViewController(title: "Friends",
                                                                        elements: elements,
                                                                        elementsPickerCompletionHandler: { (selectedIds) in
                                   let friendUuids = selectedIds as! [String]
                                   TalkApi.shared.sendCustomMessage(templateId:AppDelegate.templateId().message,
                                                                    templateArgs:["title":"제목입니다.", "description":"설명입니다."],
                                                                    receiverUuids:friendUuids) { (messageSendResult, error) in
                                       if let error = error {
                                           self?.errorHandler(error: error)
                                       }
                                       else {
                                           self?.success()
                                           
                                        //do something
                                           _ = messageSendResult
                                       }
                                   }

                               })
                               self?.navigationController?.pushViewController(vc, animated: true)
                           }
                        }
                    }
                }),
                //MARK: 톡 친구에게 send message - 스크랩메세지
                Row(title: "sendScrapMessage(requestUrl:receiverUuids:)", action: {
                    TalkApi.shared.friends { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            guard let friends = friends else {
                               self?.errorHandler(error: error)
                               return
                            }
                           
                            if let elements = self?.extractPickerElements(friends: friends) {
                                let vc = SDKSamplePickerViewController(title: "Friends",
                                                                        elements: elements,
                                                                        elementsPickerCompletionHandler: { (selectedIds) in
                                   let friendUuids = selectedIds as! [String]
                                   TalkApi.shared.sendScrapMessage(requestUrl:"https://developers.kakao.com", receiverUuids:friendUuids) { (messageSendResult, error) in
                                       if let error = error {
                                           self?.errorHandler(error: error)
                                       }
                                       else {
                                           self?.success()
                                           
                                            //do something
                                           _ = messageSendResult
                                       }
                                   }

                               })
                               self?.navigationController?.pushViewController(vc, animated: true)
                           }
                        }
                    }
                }),
                Row(title: "sendScrapMessage(requestUrl:templateId:receiverUuids:)", action: {
                    TalkApi.shared.friends { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            guard let friends = friends else {
                               self?.errorHandler(error: error)
                               return
                            }
                           
                            if let elements = self?.extractPickerElements(friends: friends) {
                                let vc = SDKSamplePickerViewController(title: "Friends",
                                                                        elements: elements,
                                                                        elementsPickerCompletionHandler: { (selectedIds) in
                                   let friendUuids = selectedIds as! [String]
                                   TalkApi.shared.sendScrapMessage(requestUrl:"https://developers.kakao.com", templateId:AppDelegate.templateId().message, receiverUuids:friendUuids) { (messageSendResult, error) in
                                       if let error = error {
                                           self?.errorHandler(error: error)
                                       }
                                       else {
                                           self?.success()
                                        
                                           //do something
                                           _ = messageSendResult
                                       }
                                   }

                               })
                               self?.navigationController?.pushViewController(vc, animated: true)
                           }
                        }
                    }
                }),
            ]),
            
            // MARK: - section FRIEND
            SectionOfRow(header: "FRIEND", items: [
                Row(title: "selectFriends(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                                viewAppearance: .auto,                              //default .auto
                                                                                orientation: .auto,                                 //default .auto
                                                                                enableSearch: true,                                 //default true
                                                                                enableIndex: true,                                  //default true
                                                                                showMyProfile: true,                                //default true
                                                                                showFavorite: true,                                 //default true
                                                                                showPickedFriend: nil,                              //default true
                                                                                maxPickableCount: nil,                              //default 30
                                                                                minPickableCount: nil)                              //default 1
                    
                    PickerApi.shared.selectFriends(params: openPickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriendsPopup(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams()
                    
                    PickerApi.shared.selectFriendsPopup(params: openPickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriend(params:)", action: {
                    
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams(title: "샘플앱 친구피커",                    //default "카카오톡 친구 선택"
                                                                    viewAppearance: .auto,                              //default .auto
                                                                    orientation: .auto,                                 //default .auto
                                                                    enableSearch: true,                                 //default true
                                                                    enableIndex: true,                                  //default true
                                                                    showMyProfile: true,                                //default true
                                                                    showFavorite: true,                                 //default true
                                                                    showPickedFriend: nil,                              //ignore
                                                                    maxPickableCount: nil,                              //ignore
                                                                    minPickableCount: nil)                              //ignore
                    
                    PickerApi.shared.selectFriend(params: openPickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriendPopup(params:)", action: {
                    let openPickerFriendRequestParams = OpenPickerFriendRequestParams()
                    
                    PickerApi.shared.selectFriendPopup(params: openPickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
            ]),
            // MARK: - section kakaoshare
            SectionOfRow(header: "KAKAOSHARE", items: [
                Row(title: "shareDefault(templatable:) - webSharer", action: {

                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        if let url = ShareApi.shared.makeDefaultUrl(templatable: templatable) {
                            print("=============================sharer default url: \(url)")
                            self.safariViewController = SFSafariViewController(url: url)
                            self.safariViewController?.modalTransitionStyle = .crossDissolve
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("웹 present success")
                            }
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:) - webSharer", action: {
                    if let url = ShareApi.shared.makeScrapUrl(requestUrl: "https://developers.kakao.com") {
                        print("=============================sharer scrap url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:templateId:) - webSharer", action: {
                    if let url = ShareApi.shared.makeScrapUrl(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link ) {
                        print("=============================sharer scrap url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareCustom(templateId:,templateArgs:) - webSharer", action: {
                    if let url = ShareApi.shared.makeCustomUrl(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."]) {
                        print("=============================sharer custom url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:)", action: {
                    ShareApi.shared.shareScrap(requestUrl: "https://developers.kakao.com") { [weak self] (sharingResult, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            guard let sharingResult = sharingResult else { return }
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                        }
                    }
                }),
                Row(title: "shareScrap(requestUrl:templateId:)", action: {
                    ShareApi.shared.shareScrap(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link) { [weak self] (sharingResult, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            guard let sharingResult = sharingResult else { return }
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                        }
                    }
                }),
                Row(title: "shareCustom(templateId:templateArgs:)", action: {
                    ShareApi.shared.shareCustom(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."]) { [weak self] (sharingResult, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            guard let sharingResult = sharingResult else { return }
                            UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                        }
                    }
                }),
                Row(title:"shareDefault(templatable:serverCallbackArgs:)", action: {

                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!
                    
                    let serverCallbackArgs = ["user_id": "abcd", "product_id": "1234"]

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable, serverCallbackArgs: serverCallbackArgs) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }

                }),
                Row(title:"shareDefault(templateObject:) using json", action: {

                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    //jsonObject
                    if let feedTemplateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templateObject: feedTemplateJsonObject) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }

                }),
                Row(title:"shareDefault(templateObject:) using constructor", action: {

                    let link = Link(webUrl: URL(string:"https://developers.kakao.com"),
                                    mobileWebUrl: URL(string:"https://developers.kakao.com"))
                    let appLink = Link(androidExecutionParams: ["key1": "value1", "key2": "value2"],
                                       iosExecutionParams: ["key1": "value1", "key2": "value2"])
                    let button1 = Button(title: "웹으로 보기", link: link)
                    let button2 = Button(title: "앱으로 보기", link: appLink)
                        
                    let itemContent = ItemContent(profileText: "Kakao",
                                    profileImageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png"),
                                    titleImageText: "Cheese cake",
                                    titleImageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png"),
                                    titleImageCategory: "cake",
                                    items: [ItemInfo(item:"cake1", itemOp:"1000원"), ItemInfo(item:"cake2", itemOp:"2000원"),
                                               ItemInfo(item:"cake3", itemOp:"3000원"), ItemInfo(item:"cake4", itemOp:"4000원"),
                                               ItemInfo(item:"cake5", itemOp:"5000원")],
                                    sum: "total",
                                    sumOp: "15000원")
                    
                    let social = Social(likeCount: 286,
                                        commentCount: 45,
                                        sharedCount: 845)
                    let content = Content(title: "딸기 치즈 케익",
                                          imageUrl: URL(string:"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png")!,
                                          description: "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                                          link: link)
                    let feedTemplate = FeedTemplate(content: content, itemContent: itemContent, social: social, buttons: [button1, button2])

                    //encode 는 이렇게 한다.
                    if let feedTemplateJsonData = (try? SdkJSONEncoder.custom.encode(feedTemplate)) {

                    //jsonObject
                        if let templateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonData) {
                            ShareApi.shared.shareDefault(templateObject:templateJsonObject) { [weak self] (sharingResult, error) in
                                if let error = error {
                                    self?.errorHandler(error: error)
                                }
                                else {
                                    self?.success()
                                    
                                    //do something
                                    guard let sharingResult = sharingResult else { return }
                                    UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                                }
                            }
                        }
                    }
                }),
                Row(title:"shareDefault(templatable:) - list", action: {

                    let listTemplateJsonStringData =
                      """
                      {
                          "object_type": "list",
                          "header_title": "WEEKLY MAGAZINE",
                          "header_link": {
                              "mobile_web_url": "https://developers.kakao.com",
                              "web_url": "https://developers.kakao.com"
                          },
                          "contents": [
                              {
                                  "title": "취미의 특징, 탁구",
                                  "description": "스포츠",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/bDPMIb/btqgeoTRQvd/49BuF1gNo6UXkdbKecx600/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "크림으로 이해하는 커피이야기",
                                  "description": "음식",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/QPeNt/btqgeSfSsCR/0QJIRuWTtkg4cYc57n8H80/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "감성이 가득한 분위기",
                                  "description": "사진",
                                  "image_url": "http://mud-kage.kakao.co.kr/dn/c7MBX4/btqgeRgWhBy/ZMLnndJFAqyUAnqu4sQHS0/kakaolink40_original.png",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ],
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }

                }),
                Row(title:"shareDefault(templatable:) - location", action: {

                    let locationTemplateJsonStringData =
                      """
                      {
                          "object_type": "location",
                          "address_title": "카카오 판교오피스 카페톡",
                          "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 8층",
                          "content": {
                              "description": "이번 주는 체리블라썸라떼 1+1",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/bSbH9w/btqgegaEDfW/vD9KKV0hEintg6bZT4v4WK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.com"
                              },
                              "title": "신메뉴 출시❤️ 체리블라썸라떼"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          }
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }

                }),
                Row(title:"shareDefault(templatable:) - commerce", action: {

                    let commerceTemplateJsonStringData =
                      """
                      {
                          "object_type": "commerce",
                          "commerce": {
                              "product_name": "Ivory long dress",
                              "regular_price": 208800,
                              "discount_price": 146160,
                              "discount_rate": 30,
                              "currency_unit": "₩",
                              "currency_unit_position": 1
                          },
                          "content": {
                              "title": "Ivory long dress (4 Color)",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/RY8ZN/btqgOGzITp3/uCM1x2xu7GNfr7NS9QvEs0/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              },
                          },
                          "buttons": [
                              {
                                  "title": "구매하기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "공유하기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }
                }),
                Row(title:"shareDefault(templatable:) - text", action: {
                    let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                    let textTemplateJsonStringData =
                    """
                    {
                        "object_type": "text",
                        "text": "\(text)",
                        "link": {
                            "web_url": "https://developers.kakao.com",
                            "mobile_web_url": "https://developers.kakao.com"
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }
                    else {
                        print("templatable is nil")
                    }
                }),
                Row(title:"shareDefault(templatable:) - calendar", action: {
                    let calendarTemplateJsonStringData =
                      """
                        {
                            "object_type": "calendar",
                            "id_type": "\(CalendarTemplate.IdType.Calendar.rawValue)",
                            "id": "\(AppDelegate.calendarId().calendar)",
                            "content": {
                                "link": {
                                    "web_url": "https://devtalk.kakao.com",
                                    "mobile_web_url": "https://developers.kakao.com"
                                },
                                "title": "2022 카타르 월드컵",
                                "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                            },
                            "buttons": [
                                {
                                    "title": "월드컵 일정 보기",
                                    "link": {
                                        "mobile_web_url": "https://developers.kakao.com",
                                        "web_url": "https://developers.kakao.com"
                                    }
                                }
                            ]
                        }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }
                }),
                Row(title:"shareDefault(templatable:) - calendar(event)", action: {
                    let calendarTemplateJsonStringData =
                      """
                      {
                          "object_type": "calendar",
                          "id_type": "\(CalendarTemplate.IdType.Event.rawValue)",
                          "id": "\(AppDelegate.calendarId().event)",
                          "content": {
                              "link": {
                                  "web_url": "https://devtalk.kakao.com",
                                  "mobile_web_url": "https://developers.kakao.com"
                              },
                              "title": "1월 신작 평론 모임",
                              "image_url": "http://k.kakaocdn.net/dn/dFUqwp/bl3SUTqb2VV/VFSqyPpKUzZVVMcmotN9A0/kakaolink40_original.png"
                          },
                          "buttons": [
                              {
                                  "title": "모임 주제 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(CalendarTemplate.self, from: calendarTemplateJsonStringData) {
                        ShareApi.shared.shareDefault(templatable: templatable) { [weak self] (sharingResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                guard let sharingResult = sharingResult else { return }
                                UIApplication.shared.open(sharingResult.url, options: [:], completionHandler: nil)
                            }
                        }
                    }
                }),
                Row(title:"imageUpload(image:)", action: {
                    if let image = UIImage(named: "sample1") {
                        ShareApi.shared.imageUpload(image:image) { [weak self] (imageUploadResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                _ = imageUploadResult
                            }
                        }
                    }
                }),
                Row(title:"imageScrap(image:)", action: {
                    if let url = URL(string:"https://t1.kakaocdn.net/kakaocorp/Service/KakaoTalk/pc/slide/talkpc_theme_01.jpg") {
                        ShareApi.shared.imageScrap(imageUrl: url, completion: { [weak self] (imageUploadResult, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success()
                                
                                //do something
                                _ = imageUploadResult
                            }
                        })
                    }
                })
            ]),
            // MARK: - section kakao navi
            SectionOfRow(header: "KAKAONAVI", items: [
                Row(title: "share(destination:) - katec", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "321286", y: "533707")
                    guard let shareUrl = NaviApi.shared.shareUrl(destination: destination) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(shareUrl) {
                        UIApplication.shared.open(shareUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "share(destination:) - wgs84", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "127.108640", y: "37.402111")
                    let option = NaviOption(coordType: .WGS84)
                    guard let shareUrl = NaviApi.shared.shareUrl(destination: destination, option: option) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(shareUrl) {
                        UIApplication.shared.open(shareUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "navigate(destination:viaList:) - katec", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "321286", y: "533707")
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: "321525", y: "532951")]
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination, viaList: viaList) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(navigateUrl) {
                        UIApplication.shared.open(navigateUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                }),
                Row(title: "navigate(destination:viaList:) - wgs84", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: "127.108640", y: "37.402111")
                    let option = NaviOption(coordType: .WGS84)
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: "127.111492", y: "37.395225")]
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination, option:option, viaList: viaList) else {
                        return
                    }
                    if UIApplication.shared.canOpenURL(navigateUrl) {
                        UIApplication.shared.open(navigateUrl, options: [:], completionHandler: nil)
                    }
                    else {
                        UIApplication.shared.open(NaviApi.webNaviInstallUrl, options: [:], completionHandler: nil)
                    }
                })
            ]),
            // MARK: - section kakaosync
            SectionOfRow(header: "KAKAOSYNC", items: [
                 Row(title: "약관선택 로그인", action: {
                    
                    let serviceTerms = ["service", "policy"]
                    
                    UserApi.shared.loginWithKakaoAccount(serviceTerms: serviceTerms, completion: {[weak self](oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            //do something
                            _ = oauthToken
                        }
                    })
                 }),
                Row(title: "약관선택 로그인 with prompt", action: {
                   
                   let serviceTerms = ["service", "policy"]
                   
                    UserApi.shared.loginWithKakaoAccount(prompts:[.Login], serviceTerms: serviceTerms, completion: {[weak self](oauthToken, error) in
                       if let error = error {
                           self?.errorHandler(error: error)
                       }
                       else {
                           self?.success()
                           
                           //do something
                           _ = oauthToken
                       }
                   })
                })
            ]),
            // MARK: - section etc
            SectionOfRow(header: "ETC", items: [
                Row(title: "paramPickerTest()", action: {                    
//                    let apiParams = ApiParams(name:"loginWithKakaoAccount", paramItems: [
//                                        ParamItem(key: "prompt", value:.multiSelectableItem(["login", "create", "cert"])),
//                                        ParamItem(key: "launchMethod", value:.singleSelectableItem(["universal", "custom_scheme"])),
//                                        ParamItem(key: "state", value:.editableText("test_state")),
//                                        ParamItem(key: "nonce", value:.editableText("test_nonce")),
//                                        ParamItem(key: "isPopup", value:.singleSelectableItem(["true", "false"])),
//                                        ParamItem(key: "scopes", value:.editableText("user,friend,chat"))
//                                    ])
//                    let vc = SDKSampleParamViewController(apiParams: apiParams,
//                                                          paramViewCompletionHandler: { [weak self] (resultApiParams) in
////                        print("resultApiParams: \(String(describing:resultApiParams))")
//
//                        self?.apiParamsInfo.info[ApiKey.loginWithKakaoTalk] = resultApiParams
//                        let paramItems = resultApiParams.paramItems
//
//                        UserApi.shared.loginWithKakaoAccount(scopes: paramItems.requestParam(key:"scopes") ?? [],
//                                                             state: paramItems.requestParam(key:"state"),
//                                                             nonce: paramItems.requestParam(key:"nonce"),
//                                                             completion:{ [weak self](oauthToken, error) in
//                            if let error = error {
//                                self?.errorHandler(error: error)
//                            }
//                            else {
//                                self?.success()
//
//                                //do something
//                                _ = oauthToken
//
//                            }
//                        })
                        
//                        let keys = paramItems.map { paramItem in return paramItem.key }
//                        paramItems.forEach { item in
//                            switch item.value {
//                            case .editableText(let _value):
//                                if (_value.contains(",")) {
//                                    let items = _value.components(separatedBy: ",")
//                                    print("key: \(item.key) value: \(items)")
//                                }
//                                else {
//                                    print("key: \(item.key) value: \(_value)")
//                                }
//                            case .multiSelectableItem(let _values):
//                                fallthrough
//                            case .singleSelectableItem(let _values):
//                                item.selectedIndex.forEach { index in
//                                    print("key: \(item.key) value: \(_values[index])")
//                                }
//                            default:
//                                break
//                            }
//                        }
//                    })
////                    let nvc = UINavigationController(rootViewController: vc)
//                    vc.modalPresentationStyle = .overFullScreen
//                    vc.modalTransitionStyle = .crossDissolve
//                    if let topVC = UIApplication.getMostTopViewController() {
//                        DispatchQueue.main.async {
//                            topVC.present(vc, animated: true) {
//                                print("presented")
//                            }
//                        }
//                    }
//                    self.navigationController?.pushViewController(vc, animated: true)
                }),
                Row(title: "shareDefault(templatable:) - webSharer", action: {
                    
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!
                    
                    let serverCallbackArgs = ["user_id": "abcd", "product_id": "1234"]

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        if let url = ShareApi.shared.makeDefaultUrl(templatable: templatable, serverCallbackArgs: serverCallbackArgs) {
                            print("=============================sharer default url: \(url)")
                            self.safariViewController = SFSafariViewController(url: url)
                            self.safariViewController?.modalTransitionStyle = .crossDissolve
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("웹 present success")
                            }
                        }
                    }
                }),
                Row(title: "Get SDK variables", action: {
                    print(">> appKey: \(try! KakaoSDK.shared.appKey())")
                    print(">> KA Header: \(KakaoSDK.shared.kaHeader())")
                    print(">> redirectUri: \(KakaoSDK.shared.redirectUri())")
                    
                }),
                
                Row(title: "Set custom token_manager ", action: {
                    print(">> Set custom token manager ")
                    Auth.shared.setTokenManager(TestCustomTokenManager())
                }),
                Row(title: "Check custom token_manager token ", action: {
                    var found = false
                    for (key, value) in UserDefaults.standard.dictionaryRepresentation() {
                        if (key == TestCustomTokenManager.tokenKey) {
                            print(">> custom token_manager token: \(String(describing:value))\n")
                            found = true
                            break
                        }
                    }
                    
                    if (found == false) {
                        print(">> not exist custom token_manager token")
                    }
                }),
                Row(title: "Set default token_manager ", action: {
                    print(">> Set default token manager ")
                    Auth.shared.setTokenManager()
                }),
                Row(title: "open WebView", action: {
                    let vc = SDKSampleWebViewController()
                    vc.modalPresentationStyle = .fullScreen
                    self.present(vc, animated: true)
                }),
                Row(title: "open WebView for JS postMessage", action: {
                    let vc = SDKSampleWebViewController(url:"https://sandbox-login.devel.kakao.com/ajax/test")
                    vc.modalPresentationStyle = .fullScreen
                    self.present(vc, animated: true)
                })
            ])
        ]
    }
}

// MARK: - custom token manager

class TestCustomTokenManager : TokenManagable {
    static let tokenKey = "test_token_key"
    
    func getToken() -> OAuthToken? {
        print(">> TestCustomTokenManager.getToken() called.")
        
        if let data = UserDefaults.standard.data(forKey: TestCustomTokenManager.tokenKey) {
            let token = try? JSONDecoder().decode(OAuthToken.self, from:data)
            print(">> loaded token: \(String(describing:token))\n\n")
            
            return token
        }
        return nil
    }
    
    func setToken(_ token: OAuthToken?) {
        print(">> TestCustomTokenManager.setToke(_) called. token: \(String(describing:token)) ")
        
        if let data = try? JSONEncoder().encode(token) {
            print(">> 토큰 암호화를 권장합니다.\n\n")
            
            UserDefaults.standard.set(data, forKey:TestCustomTokenManager.tokenKey)
            UserDefaults.standard.synchronize()
        }
    }
    
    func deleteToken() {
        print(">> TestCustomTokenManager.deleteToken() called.")
        
        UserDefaults.standard.removeObject(forKey: TestCustomTokenManager.tokenKey)
    }
}
